import java.net.*;
import java.io.*;
import java.util.StringTokenizer;

/**
 * Does the hard work of processing the interaction between
 * client and server.  Operates to the specification of 
 * the GCP 0.9 protocol specification.
 */
public class GraffitiProtocol {
	public static final String VERSION = "1.0";
	
	public boolean isConnected = true;
	
	// constants for protocol states
	private static final int START = 0;
    private static final int WAITING = 1;
    private static final int RECEIVELINE = 2;
    private static final int RECEIVEiHEADER = 3;
    private static final int TERMINAL = 4;
	private static final int SENDiHEADER = 5;
	private static final int SENDLINE = 6;
	private static final int ERROR = 7;
	
	private static final int LINE_PARAMS = 6;		// number of tokens in a LINE command + 1
	private static final int IMAGESTART_PARAMS = 4; // number of tokens in an IMAGESTART + 1
	
	// constants for protocol messages 
	// B - Message from Server or Client
	// C - Message from Client
	// S - Message from Server
	public static final String C_PUTIMAGE = "PUT IMAGE";
	public static final String C_GETNEXT = "GET NEXT";
	public static final String C_GETPREV = "GET PREVIOUS";
	public static final String C_BYE = "BYE";
	public static final String B_IMAGESTART = "IMAGESTART";
	public static final String B_LINE = "LINE";
	public static final String B_IMAGEEND = "IMAGEEND";
	public static final String S_ERROR = "ERROR";
	public static final String S_BYE = "Bye.";
	public static final String S_OK = "OK";

    private int state = START;			// initial state
	private String theInput;			// use for client's message
	private String errorMsg;			// stores current error message
	
	private static final Image BOL = new Image(-111, -111);		// BOL representation
	private static final Image EOL = new Image(-999, -999);		// EOL representation
	
	private Image currentImage = BOL;	// client starts at BOL
	private Image newImage;				// used to add new images (one per client)
	private Line currentLine;			// client's current line in currentImage
	
	private static ImgList imageList = new ImgList();		// static list of images
	
	/**
	 * Central hub for processing converstations.  This is where
	 * we end up in between states.  When one state finishs, it 
	 * enters the next state from this method.
	 * 
	 * @param in	BufferedReader input from client
	 * @param out	PrintWrite output to client
	 * 
	 * @throws	IOException	When a socket error occurs 
	 */
    public void processInput(BufferedReader in, PrintWriter out) throws IOException {
		
		errorMsg = null;
		
		if (state == START)									/*** START State ***/
		{
			processStartState(in, out);
        }
		
		if (state == WAITING)								/*** WAITING State ***/
		{
			processWaitingState(in, out);
		} 
		
		if (state == RECEIVEiHEADER)						/*** RECEIVEiHEADER State ***/
		{
			processReceiveiHeaderState(in, out);
		} 
		
		if (state == RECEIVELINE)							/*** RECEIVELINE State ***/
		{
			processReceiveLineState(in, out);
		} 
		
		if (state == SENDiHEADER)							/*** SENDiHEADER State ***/
		{
			processSendiHeaderState(out);
		} 
		
		if (state == SENDLINE)								/*** SENDLINE State ***/
		{
			processSendLineState(out);
		}
		if (state == ERROR)									/*** ERROR State ***/
		{								
			processErrorState(in, out);
		}
		
		if (state == TERMINAL)								/*** TERMINAL State ***/
		{						
			out.println(S_BYE);
			isConnected = false;
		}
    }
	
	/**
	 * Processes the client/server interaction in the
	 * START state.
	 * 
	 * @param in	BufferedReader input from client
	 * @param out	PrintWrite output to client
	 * 
	 * @throws	IOException	When a socket error occurs
	 */
	private void processStartState(BufferedReader in, PrintWriter out) throws IOException
	{
		out.println("GCP Version " + VERSION);
        state = WAITING;		
	}
	
	/**
	 * Processes the client/server interaction in the
	 * WAITING state.  Reads the message from the client
	 * and prepares the appropriate response/action.
	 * 
	 * @param in	BufferedReader input from client
	 * @param out	PrintWrite output to client
	 * 
	 * @throws	IOException	When a socket error occurs
	 */
	private void processWaitingState(BufferedReader in, PrintWriter out) throws IOException
	{
		theInput = in.readLine();
		
		if (theInput.equals(C_PUTIMAGE)) 
		{
			state = RECEIVEiHEADER;
		} 
		else if (theInput.equals(C_GETPREV)) 
		{	
			// process get previous
			if (currentImage == BOL) 
			{ 
				// do nothing 
			}
			else if (currentImage == EOL)
			{
				// currentImage gets last image if
				// it exists, otherwise BOL
				if (imageList.tail != null)
					currentImage = imageList.tail;
				else
					currentImage = BOL;
			}
			else
			{
				// currentImage gets previous image if
				// it exists, otherwise BOL
				if (currentImage.prev != null)
					currentImage = currentImage.prev;
				else
					currentImage = BOL;
			}				
			
			state = SENDiHEADER;
		} 
		else if (theInput.equals(C_GETNEXT)) 
		{
			// process get next
			if (currentImage == EOL) 
			{
				// do nothing
			}
			else if (currentImage == BOL)
			{
				// currentImage gets first image if
				// it exists, otherwise EOL
				if (imageList.tail != null)
					currentImage = imageList.head;
				else
					currentImage = EOL;
			}
			else
			{
				// currentImage gets next image if
				// it exists, otherwise EOL
				if (currentImage.next != null)
					currentImage = currentImage.next;
				else
					currentImage = EOL;
			}				
			
			state = SENDiHEADER;
		}
		else if (theInput.equals(C_BYE)) 
		{
			// end converstation
			state = TERMINAL;
		}
		else 
		{
			state = ERROR;
			errorMsg = "Unknown Command";
		}
	}

	
	/**
	 * Processes the client/server interaction in the
	 * RECEIVEiHEADER state. Reads the client header
	 * message and if everything goes well, creates 
	 * a temporary image structure with the given
	 * dimensions.
	 * 
	 * @param in	BufferedReader input from client
	 * @param out	PrintWrite output to client
	 * 
	 * @throws	IOException	When a socket error occurs
	 */
	private void processReceiveiHeaderState(BufferedReader in, PrintWriter out) throws IOException
	{
		theInput = in.readLine();
		
		String token;
		StringTokenizer tok = new StringTokenizer(theInput);
		String[] command = new String[IMAGESTART_PARAMS];
		
		int count = 0;

		while (tok.hasMoreTokens())	
		{
			command[count] = tok.nextToken();
			count++;
		}

		// verify we have the right number of tokens
		if (command[0] == null || command[1] == null 
			|| command[2] == null || command[3] != null) 
		{
			state = ERROR;
			errorMsg = "Invalid Command or Incorrect Number of Parameters";
		}
		else
		{
			if (command[0].equals(B_IMAGESTART)) {
				int width = 0, height = 0;
				
				try {
					width = Integer.parseInt(command[1]);
					height = Integer.parseInt(command[2]);
							
					// check for valid width and height, then
					// if everything is ok, create a temporary 
					// image structure and set it's width and height
					if (width < 0 || height < 0)
					{
						state = ERROR;
						errorMsg = "Width and Height must be greater than 0";
					}
					else
					{
						newImage = new Image(width, height);
						state = RECEIVELINE;				
					}
				} catch (NumberFormatException e) {
					state = ERROR;
					errorMsg = "Incorrect Header Format";
				}
			}
			else
			{
				state = ERROR;
				errorMsg = "Invalid Command";
			}					
		}
	}		

	/**
	 * Processes the client/server interaction in the
	 * RECEIVELINE state. Reads the message from the client.
	 * If it is a line (in good format), it adds the line 
	 * to the temporary image structure waiting to be inserted.
	 * If the message indicates the end of the image, the method
	 * calls a function in ImgList to add the image to 
	 * the server's list of images.
	 * 
	 * @param in	BufferedReader input from client
	 * @param out	PrintWrite output to client
	 * 
	 * @throws	IOException	When a socket error occurs
	 * 
	 * @see imgList
	 */
	private void processReceiveLineState(BufferedReader in, PrintWriter out) throws IOException
	{
		theInput = in.readLine();
		
		int count = 0;
		StringTokenizer tok = new StringTokenizer(theInput);
		String[] command = new String[LINE_PARAMS];
			
		if (theInput.equals(B_IMAGEEND))
		{
			// process image (ie. Add image to end of list)
			if (imageList.addImage(newImage) != 0) {
				state = ERROR;
				errorMsg = "Unable to add image";
			} 
			else				
			{
				//imageList.printList();	// ** DEBUGGING **
				state = WAITING;
				out.println(S_OK);
			}
		} 
		else 
		{
			// receive a line
			
			while (tok.hasMoreTokens()) {
				command[count] = tok.nextToken();
				count++;
			}
			
			// verify we have the right number of tokens
			if (command[0] == null || command[1] == null || command[2] == null 
				|| command[3] == null || command[4] == null || command[5] != null) 
			{
				state = ERROR;
				errorMsg = "Invalid Command or Incorrect Number of Parameters";
			} 
			else
			{
				if (command[0].equals("LINE")) 
				{
					int sx = 0, sy = 0, ex = 0, ey = 0;
					
					try {
						sx = Integer.parseInt(command[1]);
						sy = Integer.parseInt(command[2]);
						ex = Integer.parseInt(command[3]);
						ey = Integer.parseInt(command[4]);
						
						// we can now add the line if it fits
						// the boundary given by (0,0) (width, height)
						if (newImage.addLine(sx, sy, ex, ey) == -1)
						{
							errorMsg = "Line does not fit boundary";
							state = ERROR;
						}									
					} catch (NumberFormatException e) {
						state = ERROR;
						errorMsg = "Incorrect Line Format";
					}						
				}
				else
				{
					state = ERROR;
					errorMsg = "Invalid Command";
				}
			}				
		}	
	}
	
	/**
	 * Processes the client/server interaction in the
	 * SENDiHEADER state.  Either sends BOL, EOL, or
	 * the image header of the corresponding image.
	 * 
	 * @param in	BufferedReader input from client
	 */	
	private void processSendiHeaderState(PrintWriter out)
	{
		// check for valid image (ie. not on BOL or EOL)
		if (currentImage == BOL)
		{
			out.println("BOL");
			state = WAITING;
		}
		else if (currentImage == EOL)
		{
			out.println("EOL");
			state = WAITING;
		}
		else
		{
			// we have a valid image
			out.println(B_IMAGESTART + " " 
						+ currentImage.width + " " 
						+ currentImage.height);
			
			// set currentLine to first line of current image
			currentLine = currentImage.firstLine;
			
			state = SENDLINE;
		}
	}
	
	/**
	 * Processes the client/server interaction in the
	 * SENDLINE state.  Sends the lines, one at a time,
	 * to the client, until there are no more.  Then
	 * indiciates that it is finished sending.
	 * 
	 * @param in	BufferedReader input from client
	 */
	private void processSendLineState(PrintWriter out)
	{
		while (currentLine != null)
		{
			out.println(B_LINE + " " + currentLine.startX 
							   + " " + currentLine.startY 
							   + " " + currentLine.endX 
							   + " " + currentLine.endY);
			
			currentLine = currentLine.next;
		} 

		// no more lines
		out.println(B_IMAGEEND);
		state = WAITING;
	}
	
	/**
	 * Processes the client/server interaction in the
	 * ERROR state.  Outputs a descriptive error message
	 * if available (ie. errorMsg not null).
	 * 
	 * @param in	BufferedReader input from client
	 * @param out	PrintWrite output to client
	 */
	private void processErrorState(BufferedReader in, PrintWriter out)
	{
		if (errorMsg != null)
			out.println(S_ERROR + " " + errorMsg);
		else
			out.println(S_ERROR);
		
		state = TERMINAL;
	}
}
