/*********************************************************
 * Represents the static list of images for the server.
 * Structure: Doubly Linked List of Images
 * 
 * @see	Image
 *********************************************************/

class ImgList {
	Image head;
	Image tail;
			   
	/**
	 * Adds an image to the image list. This is the
	 * only possible place we could have race conditions thus we 
	 * have it synchronized.  See report for details.
	 * 
	 * @param	img	Image to be added to end of list
	 * @return	0 on success, -1 on failure (not implemented)
	 */	
	public synchronized int addImage(Image img) {

		img.next = null;
		
		if (head == null) {
			head = img;
			tail = img;
		} else {
			tail.next = img;
			img.prev = tail;
			tail = img;
		}
		
		return 0;
	}
	
	/**
	 * Debugging Method to print the contents of the image
	 * list to standard output.
	 */
	public void printList() {
		Image temp = head;
		while (temp != null) {
			System.out.println("width: " + temp.width + " height: " + temp.height);
			
			Line line = temp.firstLine;

			while(line != null) {
				  System.out.println("sx = " + line.startX + " sy = " + line.startY 
									 + " ex = " + line.endX + " ey = " + line.endY);
				  line = line.next;
			}
									 
			temp = temp.next;
		}
	}
}

/**
 * Represents a single image in the image list. Made up
 * of a doubly linked list of Lines.
 * 
 * @see Line
 */
class Image {
	
	public int height, width;
	public Image prev;		// previous image
	public Image next;		// next image
	public Line firstLine;		// head or list of line objects
	public Line lastLine;		// end of list of line objects
	
	/**
	 * Initializes image with width and height.
	 * 
	 * @param	w	integer width of image in pixels
	 * @param	h	integer height of image in pixels
	 */
	public Image(int w, int h) {
		height = h;
		width = w;
	}
	
	/**
	 * Adds a line to the image.  Checks to make sure
	 * it fits within the boundaries of (0,0) and
	 * (width,height).
	 * 
	 * @param	sx	start integer x coordinate
	 * @param	sy	start integer y coordinate
	 * @param	ex	end integer x coordinate
	 * @param	ey	end integer y coordinate
	 * 
	 * @return	0 on success, -1 on error
	 */
	public int addLine(int sx, int sy, int ex, int ey) {
		// check to make sure coords are within boundaries
		if (sx < 0 || sy < 0 || sx > width || sy > height 
			|| ex < 0 || ey < 0 || ex > width || ey > height)
			return -1;
		
		Line newLine = new Line(sx,sy,ex,ey);
		newLine.next = null;

		if (firstLine == null) {
			firstLine = newLine;
			lastLine = newLine;
		} else {
			lastLine.next = newLine;
			lastLine = newLine;
		}
		
		return 0;
	}
}

/**
 * Represents a single Line of an Image.
 */
class Line
{
	public int startX, startY, endX, endY;
	public Line next;

	/**
	 * Creates a new line with the given coordinates.
	 * 
	 * @param	sx	start integer x coordinate
	 * @param	sy	start integer y coordinate
	 * @param	ex	end integer x coordinate
	 * @param	ey	end integer y coordinate 
	 */
	public Line(int sx, int sy, int ex, int ey)
	{
		startX = sx; startY = sy; endX = ex; endY = ey;
	}	
}